@@ -1,7 +1,7 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2014 Inria.  All rights reserved.
- * Copyright © 2009-2011 Université Bordeaux 1
+ * Copyright © 2009-2016 Inria.  All rights reserved.
+ * Copyright © 2009-2011 Université Bordeaux
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
  */
@@ -10,6 +10,7 @@
 #include <hwloc.h>
 #include <hwloc/plugins.h>
 #include <private/private.h>
+#include <private/misc.h>
 #include <private/xml.h>
 #include <private/debug.h>
 
@@ -48,7 +49,7 @@ static int
 hwloc__nolibxml_import_next_attr(hwloc__xml_import_state_t state, char **namep, char **valuep)
 {
   hwloc__nolibxml_import_state_data_t nstate = (void*) state->data;
-  int namelen;
+  size_t namelen;
   size_t len, escaped;
   char *buffer, *value, *end;
 
@@ -116,15 +117,10 @@ hwloc__nolibxml_import_find_child(hwloc__xml_import_state_t state,
   hwloc__nolibxml_import_state_data_t nchildstate = (void*) childstate->data;
   char *buffer = nstate->tagbuffer;
   char *end;
-  int namelen;
+  size_t namelen;
 
   childstate->parent = state;
-  childstate->next_attr = state->next_attr;
-  childstate->find_child = state->find_child;
-  childstate->close_tag = state->close_tag;
-  childstate->close_child = state->close_child;
-  childstate->get_content = state->get_content;
-  childstate->close_content = state->close_content;
+  childstate->global = state->global;
 
   /* auto-closed tags have no children */
   if (nstate->closed)
@@ -245,7 +241,8 @@ hwloc__nolibxml_import_close_content(hwloc__xml_import_state_t state)
 {
   /* put back the '<' that we overwrote to 0-terminate the content */
   hwloc__nolibxml_import_state_data_t nstate = (void*) state->data;
-  *nstate->tagbuffer = '<';
+  if (!nstate->closed)
+    *nstate->tagbuffer = '<';
 }
 
 static int
@@ -274,12 +271,12 @@ hwloc_nolibxml_look_init(struct hwloc_xml_backend_data_s *bdata,
   if (strncmp(buffer, "<topology>", 10))
     goto failed;
 
-  state->next_attr = hwloc__nolibxml_import_next_attr;
-  state->find_child = hwloc__nolibxml_import_find_child;
-  state->close_tag = hwloc__nolibxml_import_close_tag;
-  state->close_child = hwloc__nolibxml_import_close_child;
-  state->get_content = hwloc__nolibxml_import_get_content;
-  state->close_content = hwloc__nolibxml_import_close_content;
+  state->global->next_attr = hwloc__nolibxml_import_next_attr;
+  state->global->find_child = hwloc__nolibxml_import_find_child;
+  state->global->close_tag = hwloc__nolibxml_import_close_tag;
+  state->global->close_child = hwloc__nolibxml_import_close_child;
+  state->global->get_content = hwloc__nolibxml_import_get_content;
+  state->global->close_content = hwloc__nolibxml_import_close_content;
   state->parent = NULL;
   nstate->closed = 0;
   nstate->tagbuffer = buffer+10;
@@ -408,17 +405,18 @@ out:
 }
 
 static int
-hwloc_nolibxml_import_diff(const char *xmlpath, const char *xmlbuffer, int xmlbuflen,
+hwloc_nolibxml_import_diff(struct hwloc__xml_import_state_s *state,
+			   const char *xmlpath, const char *xmlbuffer, int xmlbuflen,
 			   hwloc_topology_diff_t *firstdiffp, char **refnamep)
 {
-  struct hwloc__xml_import_state_s state, childstate;
-  hwloc__nolibxml_import_state_data_t nstate = (void*) state.data;
+  hwloc__nolibxml_import_state_data_t nstate = (void*) state->data;
+  struct hwloc__xml_import_state_s childstate;
   char *refname = NULL;
   char *buffer, *tmp, *tag;
   size_t buflen;
   int ret;
 
-  assert(sizeof(*nstate) <= sizeof(state.data));
+  assert(sizeof(*nstate) <= sizeof(state->data));
 
   if (xmlbuffer) {
     buffer = malloc(xmlbuflen);
@@ -442,20 +440,20 @@ hwloc_nolibxml_import_diff(const char *xmlpath, const char *xmlbuffer, int xmlbu
     tmp++;
   }
 
-  state.next_attr = hwloc__nolibxml_import_next_attr;
-  state.find_child = hwloc__nolibxml_import_find_child;
-  state.close_tag = hwloc__nolibxml_import_close_tag;
-  state.close_child = hwloc__nolibxml_import_close_child;
-  state.get_content = hwloc__nolibxml_import_get_content;
-  state.close_content = hwloc__nolibxml_import_close_content;
-  state.parent = NULL;
+  state->global->next_attr = hwloc__nolibxml_import_next_attr;
+  state->global->find_child = hwloc__nolibxml_import_find_child;
+  state->global->close_tag = hwloc__nolibxml_import_close_tag;
+  state->global->close_child = hwloc__nolibxml_import_close_child;
+  state->global->get_content = hwloc__nolibxml_import_get_content;
+  state->global->close_content = hwloc__nolibxml_import_close_content;
+  state->parent = NULL;
   nstate->closed = 0;
   nstate->tagbuffer = tmp;
   nstate->tagname = NULL;
   nstate->attrbuffer = NULL;
 
   /* find root */
-  ret = hwloc__nolibxml_import_find_child(&state, &childstate, &tag);
+  ret = hwloc__nolibxml_import_find_child(state, &childstate, &tag);
   if (ret < 0)
     goto out_with_buffer;
   if (strcmp(tag, "topologydiff"))
@@ -506,7 +504,7 @@ hwloc__nolibxml_export_update_buffer(hwloc__nolibxml_export_state_data_t ndata,
   if (res >= 0) {
     ndata->written += res;
     if (res >= (int) ndata->remaining)
-      res = ndata->remaining>0 ? ndata->remaining-1 : 0;
+      res = ndata->remaining>0 ? (int)ndata->remaining-1 : 0;
     ndata->buffer += res;
     ndata->remaining -= res;
   }
@@ -515,7 +513,7 @@ hwloc__nolibxml_export_update_buffer(hwloc__nolibxml_export_state_data_t ndata,
 static char *
 hwloc__nolibxml_export_escape_string(const char *src)
 {
-  int fulllen, sublen;
+  size_t fulllen, sublen;
   char *escaped, *dst;
 
   fulllen = strlen(src);
@@ -679,15 +677,15 @@ hwloc_nolibxml_export_buffer(hwloc_topology_t topology, char **bufferp, int *buf
 
   bufferlen = 16384; /* random guess for large enough default */
   buffer = malloc(bufferlen);
-  res = hwloc___nolibxml_prepare_export(topology, buffer, bufferlen);
+  res = hwloc___nolibxml_prepare_export(topology, buffer, (int)bufferlen);
 
   if (res > bufferlen) {
     buffer = realloc(buffer, res);
-    hwloc___nolibxml_prepare_export(topology, buffer, res);
+    hwloc___nolibxml_prepare_export(topology, buffer, (int)res);
   }
 
   *bufferp = buffer;
-  *buflenp = res;
+  *buflenp = (int)res;
   return 0;
 }
 
@@ -713,7 +711,7 @@ hwloc_nolibxml_export_file(hwloc_topology_t topology, const char *filename)
     }
   }
 
-  ret = fwrite(buffer, 1, bufferlen-1 /* don't write the ending \0 */, file);
+  ret = (int)fwrite(buffer, 1, bufferlen-1 /* don't write the ending \0 */, file);
   if (ret == bufferlen-1) {
     ret = 0;
   } else {
@@ -771,15 +769,15 @@ hwloc_nolibxml_export_diff_buffer(hwloc_topology_diff_t diff, const char *refnam
 
   bufferlen = 16384; /* random guess for large enough default */
   buffer = malloc(bufferlen);
-  res = hwloc___nolibxml_prepare_export_diff(diff, refname, buffer, bufferlen);
+  res = hwloc___nolibxml_prepare_export_diff(diff, refname, buffer, (int)bufferlen);
 
   if (res > bufferlen) {
     buffer = realloc(buffer, res);
-    hwloc___nolibxml_prepare_export_diff(diff, refname, buffer, res);
+    hwloc___nolibxml_prepare_export_diff(diff, refname, buffer, (int)res);
   }
 
   *bufferp = buffer;
-  *buflenp = res;
+  *buflenp = (int)res;
   return 0;
 }
 
@@ -805,7 +803,7 @@ hwloc_nolibxml_export_diff_file(hwloc_topology_diff_t diff, const char *refname,
     }
   }
 
-  ret = fwrite(buffer, 1, bufferlen-1 /* don't write the ending \0 */, file);
+  ret = (int)fwrite(buffer, 1, bufferlen-1 /* don't write the ending \0 */, file);
   if (ret == bufferlen-1) {
     ret = 0;
   } else {
