@@ -1,7 +1,7 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2014 Inria.  All rights reserved.
- * Copyright © 2009-2012 Université Bordeaux 1
+ * Copyright © 2009-2016 Inria.  All rights reserved.
+ * Copyright © 2009-2012, 2015 Université Bordeaux
  * Copyright © 2009-2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
  */
@@ -41,6 +41,7 @@ hwloc_obj_type_t lstopo_show_only = (hwloc_obj_type_t) -1;
 int lstopo_show_cpuset = 0;
 int lstopo_show_taskset = 0;
 int lstopo_ignore_pus = 0;
+int lstopo_collapse = 1;
 unsigned long lstopo_export_synthetic_flags = 0;
 
 char **lstopo_append_legends = NULL;
@@ -50,9 +51,6 @@ unsigned int fontsize = 10;
 unsigned int gridsize = 10;
 enum lstopo_orient_e force_orient[HWLOC_OBJ_TYPE_MAX];
 
-static int overwrite = 0;
-static int logical = -1;
-static unsigned int legend = 1;
 static unsigned int top = 0;
 
 FILE *open_output(const char *filename, int overwrite)
@@ -97,6 +95,8 @@ static hwloc_obj_t insert_task(hwloc_topology_t topology, hwloc_cpuset_t cpuset,
       fprintf(stderr, "Failed to insert process `%s' with cpuset %s\n", name, s);
     }
     free(s);
+  } else {
+    hwloc_obj_add_info(obj, "Type", "Process");
   }
 
   return obj;
@@ -231,6 +231,78 @@ static void add_process_objects(hwloc_topology_t topology)
 #endif /* HAVE_DIRENT_H */
 }
 
+static void
+lstopo_add_collapse_attributes(hwloc_topology_t topology)
+{
+  hwloc_obj_t obj, collapser = NULL;
+  unsigned collapsed = 0;
+  /* collapse identical PCI devs */
+  for(obj = hwloc_get_next_pcidev(topology, NULL); obj; obj = hwloc_get_next_pcidev(topology, obj)) {
+    if (collapser) {
+      if (!obj->arity
+	  && obj->parent == collapser->parent
+	  && obj->attr->pcidev.vendor_id == collapser->attr->pcidev.vendor_id
+	  && obj->attr->pcidev.device_id == collapser->attr->pcidev.device_id
+	  && obj->attr->pcidev.subvendor_id == collapser->attr->pcidev.subvendor_id
+	  && obj->attr->pcidev.subdevice_id == collapser->attr->pcidev.subdevice_id) {
+	/* collapse another one */
+	hwloc_obj_add_info(obj, "lstopoCollapse", "0");
+	collapsed++;
+	continue;
+      } else if (collapsed > 1) {
+	/* end this collapsing */
+	char text[10];
+	snprintf(text, sizeof(text), "%u", collapsed);
+	hwloc_obj_add_info(collapser, "lstopoCollapse", text);
+	collapser = NULL;
+	collapsed = 0;
+      }
+    }
+    if (!obj->arity) {
+      /* start a new collapsing */
+      collapser = obj;
+      collapsed = 1;
+    }
+  }
+  if (collapsed > 1) {
+    /* end this collapsing */
+    char text[10];
+    snprintf(text, sizeof(text), "%u", collapsed);
+    hwloc_obj_add_info(collapser, "lstopoCollapse", text);
+  }
+}
+
+static void
+lstopo_populate_userdata(hwloc_obj_t parent)
+{
+  hwloc_obj_t child;
+  struct lstopo_obj_userdata *save = malloc(sizeof(*save));
+
+  save->common.buffer = NULL; /* so that it is ignored on XML export */
+  save->common.next = parent->userdata;
+  save->fontsize = (unsigned) -1;
+  save->gridsize = (unsigned) -1;
+  parent->userdata = save;
+
+  for(child = parent->first_child; child; child = child->next_sibling)
+    lstopo_populate_userdata(child);
+}
+
+static void
+lstopo_destroy_userdata(hwloc_obj_t parent)
+{
+  hwloc_obj_t child;
+  struct lstopo_obj_userdata *save = parent->userdata;
+
+  if (save) {
+    parent->userdata = save->common.next;
+    free(save);
+  }
+
+  for(child = parent->first_child; child; child = child->next_sibling)
+    lstopo_destroy_userdata(child);
+}
+
 void usage(const char *name, FILE *where)
 {
   fprintf (where, "Usage: %s [ options ] ... [ filename.format ]\n\n", name);
@@ -250,7 +322,7 @@ void usage(const char *name, FILE *where)
 #endif
 		  ".\n");
 
-  fprintf (where, "Supported output file formats: console, txt, fig"
+  fprintf (where, "Supported output file formats: console, ascii, fig"
 #ifdef LSTOPO_HAVE_GRAPHICS
 #if CAIRO_HAS_PDF_SURFACE
 		  ", pdf"
@@ -291,8 +363,10 @@ void usage(const char *name, FILE *where)
   fprintf (where, "  --no-icaches          Do not show instruction caches\n");
   fprintf (where, "  --merge               Do not show levels that do not have a hierarchical\n"
                   "                        impact\n");
+  fprintf (where, "  --no-collapse         Do not collapse identical PCI devices\n");
   fprintf (where, "  --restrict <cpuset>   Restrict the topology to processors listed in <cpuset>\n");
   fprintf (where, "  --restrict binding    Restrict the topology to the current process binding\n");
+  fprintf (where, "  --restrict-flags <n>  Set the flags to be used during restrict\n");
   fprintf (where, "  --no-io               Do not show any I/O device or bridge\n");
   fprintf (where, "  --no-bridges          Do not any I/O bridge except hostbridges\n");
   fprintf (where, "  --whole-io            Show all I/O devices and bridges\n");
@@ -307,6 +381,7 @@ void usage(const char *name, FILE *where)
   fprintf (where, "  --gridsize 10         Set size of margin between elements\n");
   fprintf (where, "  --horiz[=<type,...>]  Horizontal graphical layout instead of nearly 4/3 ratio\n");
   fprintf (where, "  --vert[=<type,...>]   Vertical graphical layout instead of nearly 4/3 ratio\n");
+  fprintf (where, "  --rect[=<type,...>]   Rectangular graphical layout with nearly 4/3 ratio\n");
   fprintf (where, "  --no-legend           Remove the text legend at the bottom\n");
   fprintf (where, "  --append-legend <s>   Append a new line of text at the bottom of the legend\n");
   fprintf (where, "Miscellaneous options:\n");
@@ -320,7 +395,7 @@ enum output_format {
   LSTOPO_OUTPUT_DEFAULT,
   LSTOPO_OUTPUT_CONSOLE,
   LSTOPO_OUTPUT_SYNTHETIC,
-  LSTOPO_OUTPUT_TEXT,
+  LSTOPO_OUTPUT_ASCII,
   LSTOPO_OUTPUT_FIG,
   LSTOPO_OUTPUT_PNG,
   LSTOPO_OUTPUT_PDF,
@@ -338,8 +413,9 @@ parse_output_format(const char *name, char *callname)
     return LSTOPO_OUTPUT_CONSOLE;
   else if (!strcasecmp(name, "synthetic"))
     return LSTOPO_OUTPUT_SYNTHETIC;
-  else if (!strcasecmp(name, "txt"))
-    return LSTOPO_OUTPUT_TEXT;
+  else if (!strcasecmp(name, "ascii")
+	   || !strcasecmp(name, "txt") /* backward compat with 1.10 */)
+    return LSTOPO_OUTPUT_ASCII;
   else if (!strcasecmp(name, "fig"))
     return LSTOPO_OUTPUT_FIG;
   else if (!strcasecmp(name, "png"))
@@ -364,10 +440,10 @@ int
 main (int argc, char *argv[])
 {
   int err;
-  int verbose_mode = LSTOPO_VERBOSE_MODE_DEFAULT;
   hwloc_topology_t topology;
   const char *filename = NULL;
   unsigned long flags = HWLOC_TOPOLOGY_FLAG_IO_DEVICES | HWLOC_TOPOLOGY_FLAG_IO_BRIDGES | HWLOC_TOPOLOGY_FLAG_ICACHES;
+  unsigned long restrict_flags = 0;
   int merge = 0;
   int ignorecache = 0;
   char * callname;
@@ -375,14 +451,20 @@ main (int argc, char *argv[])
   enum hwloc_utils_input_format input_format = HWLOC_UTILS_INPUT_DEFAULT;
   enum output_format output_format = LSTOPO_OUTPUT_DEFAULT;
   char *restrictstring = NULL;
+  struct lstopo_output loutput;
   int opt;
   unsigned i;
 
+  loutput.overwrite = 0;
+  loutput.logical = -1;
+  loutput.legend = 1;
+  loutput.verbose_mode = LSTOPO_VERBOSE_MODE_DEFAULT;
+
   for(i=0; i<HWLOC_OBJ_TYPE_MAX; i++)
     force_orient[i] = LSTOPO_ORIENT_NONE;
   force_orient[HWLOC_OBJ_PU] = LSTOPO_ORIENT_HORIZ;
   force_orient[HWLOC_OBJ_CACHE] = LSTOPO_ORIENT_HORIZ;
-  force_orient[HWLOC_OBJ_NODE] = LSTOPO_ORIENT_HORIZ;
+  force_orient[HWLOC_OBJ_NUMANODE] = LSTOPO_ORIENT_HORIZ;
 
   /* enable verbose backends */
   putenv("HWLOC_XML_VERBOSE=1");
@@ -409,18 +491,18 @@ main (int argc, char *argv[])
     {
       opt = 0;
       if (!strcmp (argv[0], "-v") || !strcmp (argv[0], "--verbose")) {
-	verbose_mode++;
+	loutput.verbose_mode++;
       } else if (!strcmp (argv[0], "-s") || !strcmp (argv[0], "--silent")) {
-	verbose_mode--;
+	loutput.verbose_mode--;
       } else if (!strcmp (argv[0], "-h") || !strcmp (argv[0], "--help")) {
 	usage(callname, stdout);
         exit(EXIT_SUCCESS);
       } else if (!strcmp (argv[0], "-f") || !strcmp (argv[0], "--force"))
-	overwrite = 1;
+	loutput.overwrite = 1;
       else if (!strcmp (argv[0], "-l") || !strcmp (argv[0], "--logical"))
-	logical = 1;
+	loutput.logical = 1;
       else if (!strcmp (argv[0], "-p") || !strcmp (argv[0], "--physical"))
-	logical = 0;
+	loutput.logical = 0;
       else if (!strcmp (argv[0], "-c") || !strcmp (argv[0], "--cpuset"))
 	lstopo_show_cpuset = 1;
       else if (!strcmp (argv[0], "-C") || !strcmp (argv[0], "--cpuset-only"))
@@ -468,6 +550,8 @@ main (int argc, char *argv[])
 	flags |= HWLOC_TOPOLOGY_FLAG_WHOLE_IO;
       else if (!strcmp (argv[0], "--merge"))
 	merge = 1;
+      else if (!strcmp (argv[0], "--no-collapse"))
+	lstopo_collapse = 0;
       else if (!strcmp (argv[0], "--thissystem"))
 	flags |= HWLOC_TOPOLOGY_FLAG_IS_THISSYSTEM;
       else if (!strcmp (argv[0], "--restrict")) {
@@ -478,6 +562,14 @@ main (int argc, char *argv[])
 	restrictstring = strdup(argv[1]);
 	opt = 1;
       }
+      else if (!strcmp (argv[0], "--restrict-flags")) {
+	if (argc < 2) {
+	  usage (callname, stderr);
+	  exit(EXIT_FAILURE);
+	}
+	restrict_flags = (unsigned long) strtoull(argv[1], NULL, 0);
+	opt = 1;
+      }
       else if (!strcmp (argv[0], "--export-synthetic-flags")) {
 	if (argc < 2) {
 	  usage (callname, stderr);
@@ -492,9 +584,13 @@ main (int argc, char *argv[])
       else if (!strcmp (argv[0], "--vert"))
 	for(i=0; i<HWLOC_OBJ_TYPE_MAX; i++)
 	  force_orient[i] = LSTOPO_ORIENT_VERT;
+      else if (!strcmp (argv[0], "--rect"))
+	for(i=0; i<HWLOC_OBJ_TYPE_MAX; i++)
+	  force_orient[i] = LSTOPO_ORIENT_RECT;
       else if (!strncmp (argv[0], "--horiz=", 8)
-	       || !strncmp (argv[0], "--vert=", 7)) {
-	enum lstopo_orient_e orient = (argv[0][2] == 'h') ? LSTOPO_ORIENT_HORIZ : LSTOPO_ORIENT_VERT;
+	       || !strncmp (argv[0], "--vert=", 7)
+	       || !strncmp (argv[0], "--rect=", 7)) {
+	enum lstopo_orient_e orient = (argv[0][2] == 'h') ? LSTOPO_ORIENT_HORIZ : (argv[0][2] == 'v') ? LSTOPO_ORIENT_VERT : LSTOPO_ORIENT_RECT;
 	char *tmp = argv[0] + ((argv[0][2] == 'h') ? 8 : 7);
 	while (tmp) {
 	  char *end = strchr(tmp, ',');
@@ -528,7 +624,7 @@ main (int argc, char *argv[])
 	opt = 1;
       }
       else if (!strcmp (argv[0], "--no-legend")) {
-	legend = 0;
+	loutput.legend = 0;
       }
       else if (!strcmp (argv[0], "--append-legend")) {
 	if (argc < 2) {
@@ -555,7 +651,7 @@ main (int argc, char *argv[])
       } else if (!strcmp (argv[0], "--ps") || !strcmp (argv[0], "--top"))
         top = 1;
       else if (!strcmp (argv[0], "--version")) {
-          printf("%s %s\n", callname, VERSION);
+          printf("%s %s\n", callname, HWLOC_VERSION);
           exit(EXIT_SUCCESS);
       } else if (!strcmp (argv[0], "--output-format") || !strcmp (argv[0], "--of")) {
 	if (argc < 2) {
@@ -590,7 +686,7 @@ main (int argc, char *argv[])
     hwloc_topology_ignore_all_keep_structure(topology);
 
   if (input) {
-    err = hwloc_utils_enable_input_format(topology, input, input_format, verbose_mode > 1, callname);
+    err = hwloc_utils_enable_input_format(topology, input, &input_format, loutput.verbose_mode > 1, callname);
     if (err)
       return err;
   }
@@ -603,34 +699,6 @@ main (int argc, char *argv[])
     }
   }
 
-  err = hwloc_topology_load (topology);
-  if (err) {
-    fprintf(stderr, "hwloc_topology_load() failed (%s).\n", strerror(errno));
-    return EXIT_FAILURE;
-  }
-
-  if (top)
-    add_process_objects(topology);
-
-  if (restrictstring) {
-    hwloc_bitmap_t restrictset = hwloc_bitmap_alloc();
-    if (!strcmp (restrictstring, "binding")) {
-      if (lstopo_pid_number > 0)
-	hwloc_get_proc_cpubind(topology, lstopo_pid, restrictset, HWLOC_CPUBIND_PROCESS);
-      else
-	hwloc_get_cpubind(topology, restrictset, HWLOC_CPUBIND_PROCESS);
-    } else {
-      hwloc_bitmap_sscanf(restrictset, restrictstring);
-    }
-    err = hwloc_topology_restrict (topology, restrictset, 0);
-    if (err) {
-      perror("Restricting the topology");
-      /* fallthrough */
-    }
-    hwloc_bitmap_free(restrictset);
-    free(restrictstring);
-  }
-
   /* if the output format wasn't enforced, look at the filename */
   if (filename && output_format == LSTOPO_OUTPUT_DEFAULT) {
     if (!strcmp(filename, "-")
@@ -651,80 +719,124 @@ main (int argc, char *argv[])
   if (output_format == LSTOPO_OUTPUT_DEFAULT) {
     if (lstopo_show_cpuset
         || lstopo_show_only != (hwloc_obj_type_t)-1
-        || verbose_mode != LSTOPO_VERBOSE_MODE_DEFAULT)
+        || loutput.verbose_mode != LSTOPO_VERBOSE_MODE_DEFAULT)
       output_format = LSTOPO_OUTPUT_CONSOLE;
   }
 
-  if (logical == -1) {
-    if (output_format == LSTOPO_OUTPUT_CONSOLE)
-      logical = 1;
-    else if (output_format != LSTOPO_OUTPUT_DEFAULT)
-      logical = 0;
+  if (input_format == HWLOC_UTILS_INPUT_XML
+      && output_format == LSTOPO_OUTPUT_XML) {
+    /* must be after parsing output format and before loading the topology */
+    putenv("HWLOC_XML_USERDATA_NOT_DECODED=1");
+    hwloc_topology_set_userdata_import_callback(topology, hwloc_utils_userdata_import_cb);
+    hwloc_topology_set_userdata_export_callback(topology, hwloc_utils_userdata_export_cb);
   }
 
+  err = hwloc_topology_load (topology);
+  if (err) {
+    fprintf(stderr, "hwloc_topology_load() failed (%s).\n", strerror(errno));
+    return EXIT_FAILURE;
+  }
+
+  if (top)
+    add_process_objects(topology);
+
+  if (restrictstring) {
+    hwloc_bitmap_t restrictset = hwloc_bitmap_alloc();
+    if (!strcmp (restrictstring, "binding")) {
+      if (lstopo_pid_number > 0)
+	hwloc_get_proc_cpubind(topology, lstopo_pid, restrictset, HWLOC_CPUBIND_PROCESS);
+      else
+	hwloc_get_cpubind(topology, restrictset, HWLOC_CPUBIND_PROCESS);
+    } else {
+      hwloc_bitmap_sscanf(restrictset, restrictstring);
+    }
+    err = hwloc_topology_restrict (topology, restrictset, restrict_flags);
+    if (err) {
+      perror("Restricting the topology");
+      /* fallthrough */
+    }
+    hwloc_bitmap_free(restrictset);
+    free(restrictstring);
+  }
+
+  if (loutput.logical == -1) {
+    if (output_format == LSTOPO_OUTPUT_CONSOLE)
+      loutput.logical = 1;
+    else if (output_format != LSTOPO_OUTPUT_DEFAULT)
+      loutput.logical = 0;
+  }
+
+  loutput.topology = topology;
+  loutput.file = NULL;
+
+  lstopo_populate_userdata(hwloc_get_root_obj(topology));
+
+  if (output_format != LSTOPO_OUTPUT_XML && lstopo_collapse)
+    lstopo_add_collapse_attributes(topology);
+
   switch (output_format) {
     case LSTOPO_OUTPUT_DEFAULT:
 #ifdef LSTOPO_HAVE_GRAPHICS
 #if CAIRO_HAS_XLIB_SURFACE && defined HWLOC_HAVE_X11_KEYSYM
       if (getenv("DISPLAY")) {
-        if (logical == -1)
-          logical = 0;
-        output_x11(topology, NULL, overwrite, logical, legend, verbose_mode);
+        if (loutput.logical == -1)
+          loutput.logical = 0;
+        output_x11(&loutput, NULL);
       } else
 #endif /* CAIRO_HAS_XLIB_SURFACE */
 #ifdef HWLOC_WIN_SYS
       {
-        if (logical == -1)
-          logical = 0;
-        output_windows(topology, NULL, overwrite, logical, legend, verbose_mode);
+        if (loutput.logical == -1)
+          loutput.logical = 0;
+        output_windows(&loutput, NULL);
       }
 #endif
 #endif /* !LSTOPO_HAVE_GRAPHICS */
 #if !defined HWLOC_WIN_SYS || !defined LSTOPO_HAVE_GRAPHICS
       {
-        if (logical == -1)
-          logical = 1;
-        output_console(topology, NULL, overwrite, logical, legend, verbose_mode);
+        if (loutput.logical == -1)
+          loutput.logical = 1;
+        output_console(&loutput, NULL);
       }
 #endif
       break;
 
     case LSTOPO_OUTPUT_CONSOLE:
-      output_console(topology, filename, overwrite, logical, legend, verbose_mode);
+      output_console(&loutput, filename);
       break;
     case LSTOPO_OUTPUT_SYNTHETIC:
-      output_synthetic(topology, filename, overwrite, logical, legend, verbose_mode);
+      output_synthetic(&loutput, filename);
       break;
-    case LSTOPO_OUTPUT_TEXT:
-      output_text(topology, filename, overwrite, logical, legend, verbose_mode);
+    case LSTOPO_OUTPUT_ASCII:
+      output_ascii(&loutput, filename);
       break;
     case LSTOPO_OUTPUT_FIG:
-      output_fig(topology, filename, overwrite, logical, legend, verbose_mode);
+      output_fig(&loutput, filename);
       break;
 #ifdef LSTOPO_HAVE_GRAPHICS
 # if CAIRO_HAS_PNG_FUNCTIONS
     case LSTOPO_OUTPUT_PNG:
-      output_png(topology, filename, overwrite, logical, legend, verbose_mode);
+      output_png(&loutput, filename);
       break;
 # endif /* CAIRO_HAS_PNG_FUNCTIONS */
 # if CAIRO_HAS_PDF_SURFACE
     case LSTOPO_OUTPUT_PDF:
-      output_pdf(topology, filename, overwrite, logical, legend, verbose_mode);
+      output_pdf(&loutput, filename);
       break;
 # endif /* CAIRO_HAS_PDF_SURFACE */
 # if CAIRO_HAS_PS_SURFACE
     case LSTOPO_OUTPUT_PS:
-      output_ps(topology, filename, overwrite, logical, legend, verbose_mode);
+      output_ps(&loutput, filename);
       break;
 #endif /* CAIRO_HAS_PS_SURFACE */
 #if CAIRO_HAS_SVG_SURFACE
     case LSTOPO_OUTPUT_SVG:
-      output_svg(topology, filename, overwrite, logical, legend, verbose_mode);
+      output_svg(&loutput, filename);
       break;
 #endif /* CAIRO_HAS_SVG_SURFACE */
 #endif /* LSTOPO_HAVE_GRAPHICS */
     case LSTOPO_OUTPUT_XML:
-      output_xml(topology, filename, overwrite, logical, legend, verbose_mode);
+      output_xml(&loutput, filename);
       break;
     default:
       fprintf(stderr, "file format not supported\n");
@@ -732,6 +844,8 @@ main (int argc, char *argv[])
       exit(EXIT_FAILURE);
   }
 
+  lstopo_destroy_userdata(hwloc_get_root_obj(topology));
+  hwloc_utils_userdata_free_recursive(hwloc_get_root_obj(topology));
   hwloc_topology_destroy (topology);
 
   for(i=0; i<lstopo_append_legends_nr; i++)
